classdef Poisson2DSolver
    properties
        % Parameters
        P
        w
        a
        % Estimated variables
        T
        W_p
        D
        node_list
        b_list
        % Results
        x_list
        evalSol % values of the solution computed on the centers of the nodes.
                % when computed, it would be sad to throw it away...
    end
    methods
        function S = Poisson2DSolver(P, N_P, w, d_estimW, a)
            
            %% Initialization
            [n, ~] = size(P);
            S.P = P;
            S.T = QuadTree(P, []);
            d2 = S.T.getMaxDepth();
            S.w = w;
            S.W_p = arrayfun(@(x) W(P, S.T, d_estimW, x), (1:n)');
            depths_p = DesiredDepths(d2, S.W_p);
            S.T.loadV(P, S.W_p, depths_p, N_P);
            
            S.a = a;
            S.D = S.T.getMaxDepth();
            
            %% Cascadic Solution (Algorithm 1)
            S.node_list = S.T.getSortedLabelsByDepth();
            assert(numel(S.node_list) == (S.D + 1));
            
            
            fprintf('QuadTree computed :\n');
            fprintf('Max Depth : %d \n', S.D);
            fprintf('Number of nodes :\n');
            fprintf('%d ', cellfun(@numel, S.node_list));
            fprintf('\n');
            
            % Computation of b
            fprintf('Computing b :\n');
            S.b_list = {};
            % S.b_list contains sparse vectors :
            % S.b_list{d} is a 4^d x 1 sparse matrix,
            % containing the [V,Bi] with Bi the nodes of S.T at depth d.
            for d = 0:S.D
                i_d = d + 1;
                nodes_d = S.node_list{i_d};
                n_d = numel(nodes_d);
                b_d_ind = zeros(n_d, 2); % b_d_ind(i,:) : ind_i , val_i
                for i = 1:n_d
                    b_d_ind(i, :) = [QuadTree.getIndexFromLabel(nodes_d{i}), S.T.VDotGradBi(nodes_d{i})];
                    fprintf('2');
                end
                b_d = sparse(b_d_ind(:, 1), ones(n_d, 1), b_d_ind(:, 2), 4^d, 1);
                S.b_list = [S.b_list, {b_d}];
            end;
            fprintf('\n\n');
            
            % Computation of the x^d
            S.x_list = {};
            for d = 0:S.D
                i_d = d + 1;
                %nodes_d = S.node_list{i_d};
                %n_d = numel(nodes_d);
                %x_d = zeros(n_d, 1);
                b_d = S.b_list{i_d};
                
                for d2 = 0:(d-1)
                    i_d2 = d2 + 1;
                    Add2 = S.Ad1d2(d, d2);
                    % Remove the constraint met at coarser depths
                    b_d = b_d - Add2 * (S.x_list{i_d2});
                end;
                Add = S.Ad1d2(d, d);
                %fprintf('Rank of A(%d, %d) : %d\n', d, d, rank(full(Add)));
                %fprintf('Nonzero elements of b_%d : %d\n', d, nnz(b_d));
                % N.B. : the warning are due to the fact that Add and b
                % both have many empty rows, and columns (for Add)
                % corresponding to nodes that are not represented in the
                % tree
                x_d = Add \ b_d;
                S.x_list = [S.x_list, {x_d}];
            end;
        end
        function a = Ad1d2(S, d1, d2)
            fprintf('Computing A(%d, %d) :\n', d1, d2);
            i_d1 = d1 + 1;
            i_d2 = d2 + 1;
            nodes_1 = S.node_list{i_d1};
            nodes_2 = S.node_list{i_d2};
            n_d1 = numel(nodes_1);
            %n_d2 = numel(nodes_2);
            %a = sparse(n_d1, n_d2);
            a_indices = [];
            % we loop on the rows of our sparse matrix
            for i = 1:n_d1
                % instead of looping on every column, we only select a few
                % nodes : at most 16 if d2 < d1, at most 25 if d2 = d1
                overlap = S.T.overlappingNodesAtDepthD(d2, nodes_1{i});
                % overlap is a cell array containing the labels of the nodes
                % at depth d2 interacting (i.e. visible from) nodes_1{i},
                % the current row
                n_d2 = numel(overlap);
                assert(~( ((d2 < d1) && (n_d2 > 16)) || ((d2 == d1) && (n_d2 > 25))  ));
                for j = 1:n_d2
                    k = S.T.PoissonAij(nodes_1{i}, overlap{j}) ...
                        + S.a * 2^d1 * S.T.wPAij(S.P, S.w, nodes_1{i}, overlap{j});
                    % if k ~= 0
                    ind_i = QuadTree.getIndexFromLabel(nodes_1{i});
                    ind_j = QuadTree.getIndexFromLabel(overlap{j});
                    a_indices = [a_indices;[ind_i, ind_j, k]];
                    fprintf('1');
                    % end;
                end;
                fprintf('\n');
            end;
            a = sparse(a_indices(:,1), a_indices(:,2), a_indices(:,3), 4^d1, 4^d2);
            
            fprintf('\n\n');
        end
        function r = sol(S, x, y)
            fprintf('.');
            r = 0;
            for d = 0:S.D
                i_d = d + 1; % cell arrays are indexed 'from 1 to ..' instead of '0 to ..'
                d_nodes = S.node_list{i_d};
                d_coeff = S.x_list{i_d};
                for i = 1:numel(d_nodes)
                    r = r + S.T.getNode(d_nodes{i}).fun(x,y) * d_coeff(QuadTree.getIndexFromLabel(d_nodes{i}) , 1);
                end
            end
        end
        
        function [xi, yi, zi] = evaluateOnCenters(S)
            fprintf('Computing the solution on the node centers...\n');
            % In combination with griddata and contour, this should allow us to plot the solution efficiently
            S.evalSol = aux(S.T);
            xi = S.evalSol(:,1);
            yi = S.evalSol(:,2);
            zi = S.evalSol(:,3);
            function p = aux(n)
                c = n.center();
                p = [c(1), c(2), S.sol(c(1), c(2))];
                for i = 1:numel(n.children)
                    p = [p; aux(n.children{i})];
                end
            end
        end
        function [xq, yq, vq] = plotContour(S)
            [xi, yi, zi] = S.evaluateOnCenters();
            ech = 2^(-S.T.getMaxDepth()) / 8; % wouldn't make much sense to subsample any further
            fprintf('Interpolating the node values at scale %e...\n', ech);
            [xq,yq] = meshgrid(0:ech:1, 0:ech:1);
            
            vq = griddata(xi, yi, zi, xq, yq);
            contour(xq, yq, vq, 10);
            contour(xq, yq, vq, [0,0], 'LineWidth', 3, 'LineColor', 'r');
            
        end
    end
end









