classdef QuadTree < handle
    % a QuadTree is a recursive structure :
    %   label : list of int : [1, 3, 2], for instance
    %       the indices of children are in this order :
    %          +---+
    %          |3|4|
    %          +-+-+ 
    %          |1|2|
    %          +---+ 
    %   child : empty or QuadTree 4x1 struct
    %   points : list of indices corresponding to rows of P that lie in the
    %            node
    properties
        label  % name of the node
        children % child nodes (0 or 4)
        points % indices of the points contained in the node
        vector % the value of V attached to the node, needed for interpolation
        mycenter % center coordinated in a [0,1]^2 frame
        mylength % length of the square, 2^(-D)
    end
    methods
        %% Constructor
        function t = QuadTree(P, lab)
            
            % Initialization of the basic variables
            t.label = lab;
            t.mylength = t.compute_length();
            l = t.semilength();
            t.mycenter = t.compute_center();
            p = t.center();
            x = p(1); y = p(2);
            
            % We now fill in the nodes
            [~, d] = size(P);
            assert( (d == 2) | isempty(P) , 'P should be a n by 2 matrix');
            if isempty(P)
                t.points = [];
                t.children = {};
            else
                X = P(:, 1);
                Y = P(:, 2);
                % We could optimize this part, but in practice,
                % initializing the QuadTree takes very little time
                t.points = find( x-l <= X & X < x+l & y - l <= Y & Y < y+l);
                if length(t.points) >= 2
                    for i = 1:4
                        t.children{i} = QuadTree(P, [lab, i]);
                    end
                end;
            end
            
            % t.vector will be filled in by T.loadV
            t.vector = zeros(1, 2);
        end
        
        %% The node function
        function r = fun(obj, x,y)
            % The function associated to the node, as explained in the
            % report : approximates a gaussian centered on the node, of
            % variance mylength/6
            r = F((x - obj.mycenter(1))/obj.mylength, (y - obj.mycenter(2))/obj.mylength) / (obj.mylength)^2;
        end
        
        %% Small tools and utilities
        function r = depth(obj)
            % depth of the current node in the standard QuadTree
            r = length(obj.label);
        end
        function l = length(obj)
            l = obj.mylength;
        end
        function l = semilength(obj)
            l = obj.mylength / 2;
        end
        function l = compute_length(obj)
            l = (0.5^(obj.depth()));
        end
        function d = getMaxDepth(obj)
            % returns the maximal depth of the nodes in the tree
            if numel(obj.children) == 0
                d = 0;
            else
                c = cellfun(@getMaxDepth, obj.children);
                %disp(c);
                %c = cell2mat(c);
                d = 1 + max(c);
            end
        end
        function p = center(obj)
            p = obj.mycenter;
        end
        function p = compute_center(obj)
            % p = [x,y]
            %disp(obj.label);
            p = center_aux(obj.label);
            function p = center_aux(l)
                if numel(l) <= 0
                    p = [0.5,0.5];
                else
                    l1 = l(1);
                    p = ( [mod(l1 - 1, 2), l1 >= 3]/2 + center_aux(l(2:end))/2 );
                end;
            end
        end
        
        
        %% getNode, getNodes & Co.
        function n = getNode(obj, lab)
            % returns a pointer (as QuadTree < handle)
            % to the node pointed by lab
            if isempty(lab)
                n = obj;
            else
                if isempty(obj.children)
                    error('Trying to access an undefined node');
                else
                    l1 = lab(1);
                    n = getNode(obj.children{l1}, lab(2:end));
                end
            end
        end
        function n = getRepresentingNode(obj, lab)
            % this is basically getNode that returns the finer
            % result possible instead of an error if the required
            % node does not exist in the tree
            if isempty(lab) || isempty(obj.children)
                n = obj;
            else
                l1 = lab(1);
                n = getRepresentingNode(obj.children{l1}, lab(2:end));
            end
        end
        function l = getRepresentingNodeLabel(obj, lab)
            % this is basically getNode + getLabel that returns the finer
            % result possible instead of an error if the required
            % node does not exist in the tree
            if isempty(lab) || isempty(obj.children)
                l = [];
            else
                l1 = lab(1);
                l = [l1, getRepresentingNodeLabel(obj.children{l1}, lab(2:end))];
            end
        end
        function l = getSortedLabels(obj)
            % returns the labels, in "military" order
            l = {[]};
            q = {obj};
            while ~isempty(q)
                n = q{1};
                l = [l, n.label];
                if ~isempty(n.children)
                    cellfun(@aux, n.children)
                end
                q = q(2:end);
            end;
            function aux(t)
                q = [q, {t}];
            end
        end
        function l = getSortedLabelsByDepth(obj)
            % same as getSortedLabels, but the nodes are sorted by depth in
            % a Cell Array
            l1 = obj.getSortedLabels();
            n = numel(l1);
            t = 1;
            l_ref = 0;
            l = {};
            for i = 1:n
                if numel(l1{i}) > l_ref
                    l = [l, {l1(t:(i-1))}];
                    t = i;
                    l_ref = numel(l1{i});
                end;
            end;
            l = [l, {l1(t:end)}];
        end
        function c = getSortedCenters(obj)
            % same as getSortedLabels, but returns the centers instead
            l = obj.getSortedLabels();
            c = cellfun(@QuadTree.getCenterFromLabel, l, 'UniformOutput', 0);
            c = cell2mat(c');
        end
        
        %% Plotting functions
        function plot(obj)
            if ~isempty(obj.children)
                l = obj.semilength();
                p = obj.center();
                x = p(1);
                y = p(2);
                plot([x - l, x + l], [y, y], [x, x], [y-l, y+l]);
            end;
            if ~isempty(obj.children)
                cellfun(@(f) f.plot(), obj.children)
            end;
        end
        
        function plotV(obj)
            % quiver the accumulated V vectors on each node center
            [X, Y, U, V] = aux(obj);
            quiver(X, Y, U, V, 5);
            
            function [x, y, u, v] = aux(node)
                p = node.center();
                x = p(1);   y = p(2);
                vec = node.vector;
                u = vec(1); v = vec(2);
                
                if ~isempty(node.children)
                    n = numel(node.children);
                    for t = 1:n
                        [xc, yc, uc, vc] = aux(node.children{t});
                        x = [x;xc];
                        y = [y;yc];
                        u = [u;uc];
                        v = [v;vc];
                    end;
                end;
            end
        end
        
        
        
        
        %% Computation of V
        
        function [] = loadV(obj, P, W, D, V)
            % W is the vector of local densities
            % D is the vector of desired depths (i.e. -log2(density/mean-density))
            % V are the local inward pointing normals
            [n, ~] = size(W);
            % For each point p_i of P, we accumulate V_i on its neighbors 
            for i = 1:n
                [o_i, a_i] = QuadTree.getBarycentricCoordinates(P(i, :), D(i));
                [n_n, ~] = size(o_i); % n_n the number of neighbours of p_i
                for k = 1:n_n
                    obj.addToVFromAbsCoordinates(o_i(k, :), a_i(k) * V(i, :) / W(i));
                end;
            end;
        end
        
        function [] = addToVFromAbsCoordinates(obj, coord, val)
            aux(obj, coord);
            % as we are dealing with double precision float for the
            % centers, we have to make sure the depth is not TOO large.
            % Obviously, the depth 100 chosen there is completely
            % arbitrary : according to the matlab doc, this could more or
            % less be about 300.
            assert(isequal(mod(coord*(2^100), 1), [0,0]));
            function nt = aux(t, c)
                nt = t;
                if isequal(c, [0.5, 0.5]) % i.e. we are on the designated node
                    nt.vector = nt.vector + val; % we accumulate the value
                else
                    x = (c(1) >= 0.5) + 1;
                    y = (c(2) >= 0.5) + 1;
                    child = nt.getOrCreateChildren(x + 2*(y-1));
                    child = aux(child, mod(2*c, 1));
                    nt.children{x + 2*(y-1)} = child;
                end
            end
        end
        
        function c = getOrCreateChildren(obj, ind)
            if isempty(obj.children)
                % We create empty nodes on purpose
                for i = 1:4
                    obj.children{i} = QuadTree([], [obj.label, i]);
                end;
                c = getOrCreateChildren(obj, ind);
            else
                c = obj.children{ind};
            end
        end
        
        
        function r = VDotGradBi(obj, lab)
            % returns the scalar product between V and grad Bi when Bi is
            % the node of center c
            % N.B. : in the current implementation, we do not restrict the
            % computation to [0,1]^2
            c = QuadTree.getCenterFromLabel(lab);
            depth = numel(lab);
            leng  = 2^(-depth);
            
            r = aux(obj);
            
            function s = aux(node)
                % main idea is that, if n does not interfer with m, then
                % neither do the children of n
                if QuadTree.funOverlap(node.label, lab)
                    s = VNodeDotGradBi(node.center(), node.length(), node.vector);
                    n_c = numel(node.children);
                    if n_c > 0
                        for t = 1:n_c
                            s = s + aux(node.children{t});
                        end;
                    end
                else
                    s = 0;
                end;
            end
            function s = VNodeDotGradBi(cnode, lnode, vnode)
                s = (vnode(1) * FDotGradFx((cnode - c ) / leng, lnode/leng ) ...
                    + vnode(2) * FDotGradFy((cnode - c ) / leng, lnode/leng )) / leng^3; % leng^(dim + 1), actually
            end
        end
        
        %% Computation of A
        function L = overlappingNodesAtDepthD(obj, d, lab)
            % returns the labels of the nodes of obj,
            % at depth d, which interact with the node 'lab'
            centers = QuadTree.theoreticalOverlappingNodesAtDepthD(d, lab);
            [n, ~] = size(centers);
            L = {};
            for i = 1:n
               try
                   lbl = aux(centers(i, :));
                   L = [L, {lbl}];
               catch ME
                   if ~(strcmp(ME.identifier,'QuadTree:noNode'))
                       rethrow(ME);
                   end;
               end;
            end
            
            %L = cellfun(@aux, C);
            
            function lbl = aux(c)
                lbl = aux2(obj, c);
                function l = aux2(node, center)
                    if isequal(center, [0.5, 0.5])
                        l = [];
                    else
                        if ~isempty(node.children)
                            x = floor(center(1) * 2) + 1;
                            y = floor(center(2) * 2) + 1;
                            ind = x + 2*(y-1);
                            l = [ind, aux2(node.children{ind}, mod(2*center, 1))];
                        else
                            errorStruct.message = 'No need to worry...';
                            errorStruct.identifier = 'QuadTree:noNode';
                            error(errorStruct); 
                        end
                    end;
                end
            end
            
        end
        
        
        
        function aij = PoissonAij(obj, lab_i, lab_j)
            ci = QuadTree.getCenterFromLabel(lab_i);
            depth_i = numel(lab_i);
            leng_i  = 2^(-depth_i);
            cj = QuadTree.getCenterFromLabel(lab_j);
            depth_j = numel(lab_j);
            leng_j  = 2^(-depth_j);
            aij = GradBiDotGradBj(ci, leng_i, cj, leng_j);
        end
        
        function aij = wPAij(obj, P, w, lab_i, lab_j)
            % computes [Bi,Bj]_(w,P)
            % depth_j <= depth_i with algorithm 1 anyway...
            l_ij = obj.getOverlappingPoints(lab_j, lab_i);
            
            if isempty(l_ij)
                aij = 0;
            else
                p_ij = P(l_ij, :);
                n_i = obj.getNode(lab_i);
                n_j = obj.getNode(lab_j);
                f_ij = arrayfun(@n_i.fun, p_ij(:,1), p_ij(:,2)) ;
                g_ij = arrayfun(@n_j.fun, p_ij(:,1), p_ij(:,2)) ;
                w_ij = w(l_ij);
                aij = sum(w_ij .* f_ij .* g_ij);
                aij = aij ./ mean(w_ij); % !!!
                assert(isequal(size(aij), [1, 1]));
            end
            
        end
        
        function L = getOverlappingNodes(obj, lab1, lab2)
            % l1 and l2 being the labels of two nodes,
            % depth(l1) <= depth(l2)
            % L is a list containing the theoretical labels of the nodes
            % representing the intersection of supports of 'l1.fun' and 'l2.fun'
            % the nodes are taken in obj, and have as big a depth as
            % possible, <= depth(l2)
            Lu = QuadTree.getOverlappingNodesAtUniformDepth(lab1, lab2);
            L = cell(size(Lu));
            for i = 1:numel(Lu)
                % as cellfun won't work with quadtrees, we have to make a
                % loop
                L{i} = obj.getRepresentingNode(Lu{i});
            end
        end
        function L_p = getOverlappingPoints(obj, lab1, lab2)
            % l1 and l2 being the labels of two nodes,
            % depth(l1) <= depth(l2)
            % L is a vector containing the indices of the points of P
            % contained in the intersection of the two supports
            L = obj.getOverlappingNodes(lab1, lab2);
            L_p = [];
            for i = 1:numel(L)
                L_p = [L_p; L{i}.points];
            end;
            L_p = unique(L_p);
        end
    end
    
    methods(Static)
        %% Change between the different systems of coordinates :
        %  label, center, and (depth, index)
        function [o, a] = getBarycentricCoordinates(p, d)
            % p is a [1, 2] vector, the coordinates of the point
            % d is the depth we're working at
            % o is a [n, 2] matrix, with n = 1, 2 or 4,
            %    containing the centers of the nodes
            % a are the barycentric coeff
            p_i = (2^d) * p + 0.5;
            m = 2^d;
            if d <= 0
                o = [0.5, 0.5];
                a = [1];
            else
                % matlab does not allow to switch on [b1, b2], so we have
                % to use this little trick...
                switch (p_i(1) > 1) + (p_i(1) >= m ) + 3 * (p_i(2) > 1) + (p_i(2) >= m )
                    case 4 %[1, 1] % By far the most frequent case
                        o = zeros(4, 2);
                        a = zeros(4, 1);
                        f = floor(p_i);
                        o(3, :) = [f(1), f(2) + 1];  o(4, :) = [f(1) + 1, f(2) + 1];
                        o(1, :) = [f(1), f(2)];      o(2, :) = [f(1) + 1, f(2)];
                        xy = p_i - f;
                        x = xy(1);   y = xy(2);
                        a(3) = (1-x)*y;      a(4) = x*y;
                        a(1) = (1-x)*(1-y);  a(2) = x*(1-y);
                        
                        
                    case 3 %[0, 1] % Left Side
                        o = zeros(2, 2);
                        a = zeros(2, 1);
                        f = floor(p_i(2));
                        o(2,:) = [1, f + 1];
                        o(1,:) = [1, f];
                        y = p_i(2) - f;
                        a(2) = (1-y);
                        a(1) = y;
                    case 5 %[2, 1] % Right Side
                        o = zeros(2, 2);
                        a = zeros(2, 1);
                        f = floor(p_i(2));
                        o(2,:) = [m, f + 1];
                        o(1,:) = [m, f];
                        y = p_i(2) - f;
                        a(2) = (1-y);
                        a(1) = y;
                        
                    case 1 %[1, 0] % Bottom Side
                        o = zeros(2, 2);
                        a = zeros(2, 1);
                        f = floor(p_i(1));
                        o(2,:) = [f + 1, 1];
                        o(1,:) = [f, 1];
                        x = p_i(1) - f;
                        a(2) = (1-x);
                        a(1) = x;
                    case 7 % [1, 2] % Top Side
                        o = zeros(2, 2);
                        a = zeros(2, 1);
                        f = floor(p_i(1));
                        o(2,:) = [f + 1, m];
                        o(1,:) = [f, m];
                        x = p_i(1) - f;
                        a(2) = (1-x);
                        a(1) = x;
                        
                    case 0 %[0, 0] % Corners
                        o = [1, 1];  a = 1;
                    case 6 %[0, 2]
                        o = [1, m];  a = 1;
                    case 2 %[2, 0]
                        o = [m, 1];  a = 1;
                    case 8 %[2, 2]
                        o = [m, m];  a = 1;
                    otherwise
                        error('This should not have happened...');
                end;
                o = (o - 0.5)/2^d;
                if ~(isequal(mod(2^100*o, 1), zeros(size(o )) ))
                    error('Profondeur trop importante pour etre correctement geree');
                end;
            end;
        end
        
        function l = getLabelFromCenter(c)
            if isequal(c, [0.5, 0.5])
                l = [];
            else
                x = (c(1) >= 0.5) + 1;
                y = (c(2) >= 0.5) + 1;
                l = [x + 2*(y-1), QuadTree.getLabelFromCenter(mod(2*c, 1))];
            end
        end
        function c = getCenterFromLabel(l)
            if numel(l) == 0
                c = [0.5, 0.5];
            else
                l1 = l(1);
                x = mod(l1+1, 2)/2;
                y = (l1 >= 3)/2;
                cs = QuadTree.getCenterFromLabel(l(2:end));
                c = [x, y] + cs/2;
            end
        end
        
        function ind = getIndexFromLabel(lab)
            d = length(lab);
            x = aux(mod(lab + 1, 2));
            y = aux(lab >= 3);
            ind = x + y * 2^d + 1;
            function i = aux(l)
                if isempty(l)
                    i = 0;
                else
                    i = 2 * aux(l(1:(end-1))) + l(end);
                end;
            end
        end
        
        %% Overlap computations
        function L = getOverlappingNodesAtUniformDepth(lab1, lab2)
            % l1 and l2 being the labels of two nodes,
            % depth(l1) <= depth(l2)
            % L is a list containing the theoretical labels of the nodes at
            % depth = depth(l2) in the support of 'l1.fun' and 'l2.fun'
            
            c1 = QuadTree.getCenterFromLabel(lab1);
            c2 = QuadTree.getCenterFromLabel(lab2);
            d1 = numel(lab1);  d2 = numel(lab2);
            l1 = 2^(-d1);      l2 = 2^(-d2);
            assert(d1 <= d2);
            
            % we compute the centers of the neighboring node
            offsets = l2 * ...
                [[-1,-1];[0,-1];[1,-1];...
                [-1,0] ;[0,0] ;[1,0];...
                [-1,1] ;[0,1] ;[1,1]];
            neigh_2 = repmat(c2, 9, 1) + offsets;
            
            % the limits of the square support of 'l1.fun'
            lim_l = c1(1) - 1.5*l1;   lim_r = c1(1) + 1.5*l1;
            lim_b = c1(2) - 1.5*l1;   lim_t = c1(2) + 1.5*l1;
            
            lim_l = limits(lim_l);    lim_r = limits(lim_r);
            lim_b = limits(lim_b);    lim_t = limits(lim_t);
            
            % chosen_nodes is a boolean [9,1] vector
            chosen_nodes = (lim_l <= neigh_2(:,1)) ...
                & (lim_r >= neigh_2(:,1)) ...
                & (lim_b <= neigh_2(:,2)) ...
                & (lim_t >= neigh_2(:,2));
            chosen_centers = neigh_2(chosen_nodes, :);
            c = num2cell(chosen_centers, 2);
            L = cellfun(@QuadTree.getLabelFromCenter, c, 'UniformOutput', 0);
            
            function x = limits(x)
                x = max(min(x, 1), 0);
            end
        end
        function b = funOverlap(lab_1, lab_2)
            % returns true if the two functions have an overlapping support
            c_1 = QuadTree.getCenterFromLabel(lab_1);
            c_2 = QuadTree.getCenterFromLabel(lab_2);
            depth_1 = numel(lab_1);     depth_2 = numel(lab_2);
            leng_1  = 2^(-depth_1);     leng_2  = 2^(-depth_2);
            lref = 1.5 * (leng_1 + leng_2);
            t = abs(c_2 - c_1) < lref;
            b = t(1) & t(2);
        end
        
        
        function L = theoreticalOverlappingNodesAtDepthD(d, lab)
            % returns the center of nodes at depth d visible by the
            % node given by lab
            c = QuadTree.getCenterFromLabel(lab);
            depth = numel(lab);
            leng  = 2^(-depth);
            leng2 = 2^(-d) / 2;
            assert(d <= depth);
            if isequal(d, depth)
                [X,Y] = meshgrid(-2:1:2, -2:1:2);
                offsets = [X(:), Y(:)];
                offsets = offsets * leng;
                center_list = repmat(c, 25, 1) + offsets;
                
            else
                square = [[-1,-1];...
                           [1,-1];...
                           [-1,1];...
                           [1,1]];
                offsets = [repmat([-1, -1], 4, 1) * (leng2 + leng) + leng2 * square;...
                    repmat([1, -1], 4, 1) * (leng2 + leng) + leng2 * square;...
                    repmat([-1, 1], 4, 1) * (leng2 + leng) + leng2 * square;...
                    repmat([1,  1], 4, 1) * (leng2 + leng) + leng2 * square];  
                center_list = repmat(c, 16, 1) + offsets;
                
            end;
            inthesquare = (center_list(:,1) >= 0) ...
                        & (center_list(:,1) <= 1) ...
                        & (center_list(:,2) >= 0) ...
                        & (center_list(:,2) <= 1);
            center_list = center_list(inthesquare, :);
            
            L = center_list / (2^(-d));
            L = round(L + 0.5) - 0.5;
            L = L * (2^(-d));
            L = unique(L, 'rows');
        end
        
    end
    
end